package ai;

import java.util.HashMap;
import java.util.Map;

public class Evaluation {

	// define the basic value of each piece
	public static final int BASEVALUE_PAWN = 100;
	public static final int BASEVALUE_ADVISOR = 250;
	public static final int BASEVALUE_ELEPHANT = 250;
	public static final int BASEVALUE_ROOK = 500;
	public static final int BASEVALUE_HORSE = 350;
	public static final int BASEVALUE_CANNON = 350;
	public static final int BASEVALUE_KING = 10000;

	// define the flexibility of each piece
	public static final int FLEXIBILITY_PAWN = 15;
	public static final int FLEXIBILITY_ADVISOR = 1;
	public static final int FLEXIBILITY_ELEPHANT = 1;
	public static final int FLEXIBILITY_ROOK = 6;
	public static final int FLEXIBILITY_HORSE = 12;
	public static final int FLEXIBILITY_CANNON = 6;
	public static final int FLEXIBILITY_KING = 0;

	Map<Integer, Integer> m_BaseValue = new HashMap<Integer, Integer>();// A array to store the information of basic
																		// value of pieces
	Map<Integer, Integer> m_FlexValue = new HashMap<Integer, Integer>();// A array to store the information of
																		// flexibility of pieces
	private int[][] m_AttackPos = new int[10][9];// Store information of piece threatened at each position
	private int[][] m_GuardPos = new int[10][9];// Store information of piece protected at each position
	private int[][] m_FlexibilityPos = new int[10][9];// Store the flexibility score of piece at each position
	private int[][] m_chessValue = new int[10][9];// Store the total score of piece at each position
	private int nPosCount;// Record the number of relevant positions of a piece
	private int[][] RelatePos = new int[20][2];// An array to record the relevant positions of a piece

	// Extra value matrix of red pawn
	private static final int[][] BA0 = { { 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 90, 90, 110, 120, 120, 120, 110, 90, 90 },
			{ 90, 90, 110, 120, 120, 120, 110, 90, 90 }, { 70, 90, 110, 110, 110, 110, 110, 90, 70 },
			{ 70, 70, 70, 70, 70, 70, 70, 70, 70 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 } };

	// Extra value matrix of black pawn
	private static final int[][] BA1 = { { 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
			{ 70, 70, 70, 70, 70, 70, 70, 70, 70 }, { 70, 90, 110, 110, 110, 110, 110, 90, 70 },
			{ 90, 90, 110, 120, 120, 120, 110, 90, 90 }, { 90, 90, 110, 120, 120, 120, 110, 90, 90 },
			{ 0, 0, 0, 0, 0, 0, 0, 0, 0 } };

	public Evaluation() {
		super();
		m_BaseValue.put(0, 0);
		// Initial the array of basic value of each piece
		m_BaseValue.put(5, BASEVALUE_KING);
		m_BaseValue.put(1, BASEVALUE_ROOK);
		m_BaseValue.put(2, BASEVALUE_HORSE);
		m_BaseValue.put(4, BASEVALUE_ADVISOR);
		m_BaseValue.put(3, BASEVALUE_ELEPHANT);
		m_BaseValue.put(6, BASEVALUE_CANNON);
		m_BaseValue.put(7, BASEVALUE_PAWN);

		m_BaseValue.put(55, BASEVALUE_KING);
		m_BaseValue.put(11, BASEVALUE_ROOK);
		m_BaseValue.put(22, BASEVALUE_HORSE);
		m_BaseValue.put(44, BASEVALUE_ADVISOR);
		m_BaseValue.put(33, BASEVALUE_ELEPHANT);
		m_BaseValue.put(66, BASEVALUE_CANNON);
		m_BaseValue.put(77, BASEVALUE_PAWN);

		m_FlexValue.put(0, 0);
		// Initial the array of flexibility of each piece
		m_FlexValue.put(5, FLEXIBILITY_KING);
		m_FlexValue.put(1, FLEXIBILITY_ROOK);
		m_FlexValue.put(2, FLEXIBILITY_HORSE);
		m_FlexValue.put(4, FLEXIBILITY_ADVISOR);
		m_FlexValue.put(3, FLEXIBILITY_ELEPHANT);
		m_FlexValue.put(6, FLEXIBILITY_CANNON);
		m_FlexValue.put(7, FLEXIBILITY_PAWN);

		m_FlexValue.put(55, FLEXIBILITY_KING);
		m_FlexValue.put(11, FLEXIBILITY_ROOK);
		m_FlexValue.put(22, FLEXIBILITY_HORSE);
		m_FlexValue.put(44, FLEXIBILITY_ADVISOR);
		m_FlexValue.put(33, FLEXIBILITY_ELEPHANT);
		m_FlexValue.put(66, FLEXIBILITY_CANNON);
		m_FlexValue.put(77, FLEXIBILITY_PAWN);

		for (int i = 0; i < 10; i++) {
			for (int j = 0; j < 9; j++) {
				m_AttackPos[i][j] = 0;
				m_GuardPos[i][j] = 0;
				m_FlexibilityPos[i][j] = 0;
				m_chessValue[i][j] = 0;
			}
		}
	}

	private int count = 0; // This is a global variable used to count the number of children nodes that
							// call the valuation function

	// return the extra value of each Pawn
	private int getBingValue(int x, int y, int[][] CurSituation) {
		if (CurSituation[y][x] == 77) {// if red pawn
			return BA0[y][x];
		}
		if (CurSituation[y][x] == 7) {// if black pawn
			return BA1[y][x];
		}
		return 0;// if not pawn, return 0
	}

	// Evaluation function
	// bIsRedTurn is a sign to get which player turn, true is red, false is black
	public int evaluate(int[][] position, boolean bIsRedTurn) {
		int i, j, k;
		int nChessType, nTargetType;
		count++;

		for (i = 0; i < 10; i++) {
			for (j = 0; j < 9; j++) {
				m_AttackPos[i][j] = 0;
				m_GuardPos[i][j] = 0;
				m_FlexibilityPos[i][j] = 0;
				m_chessValue[i][j] = 0;
			}
		}

		for (i = 0; i < 10; i++)
			for (j = 0; j < 9; j++) {
				if (position[i][j] != 0) {
					nChessType = position[i][j];
					getRelatePiece(position, j, i);
					for (k = 0; k < nPosCount; k++) {
						nTargetType = position[RelatePos[k][1]][RelatePos[k][0]];
						if (nTargetType == 0) {
							m_FlexibilityPos[i][j]++;
						} else {
							if (IsSameSide(nChessType, nTargetType)) {
								m_GuardPos[RelatePos[k][1]][RelatePos[k][0]]++;
							} else {
								m_AttackPos[RelatePos[k][1]][RelatePos[k][0]]++;
								m_FlexibilityPos[i][j]++;
								switch (nTargetType) {
								case 55:
									if (!bIsRedTurn)
										return 18888;
									break;
								case 5:
									if (bIsRedTurn)
										return 18888;
									break;
								default:
									m_AttackPos[RelatePos[k][1]][RelatePos[k][0]] += (30
											+ (m_BaseValue.get(nTargetType) - m_BaseValue.get(nChessType)) / 10) / 10;
									break;
								}
							}
						}
					}
				}
			}

		for (i = 0; i < 10; i++)
			for (j = 0; j < 9; j++) {
				if (position[i][j] != 0) {
					nChessType = position[i][j];
					m_chessValue[i][j]++;
					m_chessValue[i][j] += m_FlexValue.get(nChessType) * m_FlexibilityPos[i][j];
					m_chessValue[i][j] += getBingValue(j, i, position);
				}
			}
		int nHalfvalue;
		for (i = 0; i < 10; i++)
			for (j = 0; j < 9; j++) {
				if (position[i][j] != 0) {
					nChessType = position[i][j];
					nHalfvalue = m_BaseValue.get(nChessType) / 16;
					m_chessValue[i][j] += m_BaseValue.get(nChessType);

					if (nChessType > 10) {
						if (m_AttackPos[i][j] != 0 ? true : false) {
							if (bIsRedTurn) {
								if (nChessType == 55) {
									m_chessValue[i][j] -= 20;
								} else {
									m_chessValue[i][j] -= nHalfvalue * 2;
									if (m_GuardPos[i][j] != 0 ? true : false)
										m_chessValue[i][j] += nHalfvalue;
								}
							} else {
								if (nChessType == 55)
									return 18888;
								m_chessValue[i][j] -= nHalfvalue * 10;
								if (m_GuardPos[i][j] != 0 ? true : false)
									m_chessValue[i][j] += nHalfvalue * 9;
							}
							m_chessValue[i][j] -= m_AttackPos[i][j];
						} else {
							if (m_GuardPos[i][j] != 0 ? true : false)
								m_chessValue[i][j] += 5;
						}
					} else {
						if (m_AttackPos[i][j] != 0 ? true : false) {
							if (!bIsRedTurn) {
								if (nChessType == 5) {
									m_chessValue[i][j] -= 20;
								} else {
									m_chessValue[i][j] -= nHalfvalue * 2;
									if (m_GuardPos[i][j] != 0 ? true : false)
										m_chessValue[i][j] += nHalfvalue;
								}
							} else {
								if (nChessType == 5)
									return 18888;
								m_chessValue[i][j] -= nHalfvalue * 10;
								if (m_GuardPos[i][j] != 0 ? true : false)
									m_chessValue[i][j] += nHalfvalue * 9;
							}
							m_chessValue[i][j] -= m_AttackPos[i][j];
						} else {
							if (m_GuardPos[i][j] != 0 ? true : false)
								m_chessValue[i][j] += 5;
						}
					}
				}
			}

		int nRedValue = 0;
		int nBlackValue = 0;

		for (i = 0; i < 10; i++)
			for (j = 0; j < 9; j++) {
				nChessType = position[i][j];
				// if (nChessType == R_KING || nChessType == B_KING)
				// m_chessValue[i][j] = 10000;
				if (nChessType != 0) {
					if (nChessType > 10) {
						nRedValue += m_chessValue[i][j];
					} else {
						nBlackValue += m_chessValue[i][j];
					}
				}
			}
		if (bIsRedTurn) {
			return nRedValue - nBlackValue;
		} else {
			return nBlackValue - nRedValue;
		}
	}

	// add a position into the array RelatePos
	private void AddPoint(int x, int y) {
		RelatePos[nPosCount][0] = x;
		RelatePos[nPosCount][1] = y;
		nPosCount++;
	}

	// Enumerate all the relevant positions of Pieces including accessible positions
	// and protected positions
	private int getRelatePiece(int[][] position, int j, int i) {
		nPosCount = 0;
		int nChessID;
		boolean flag;
		int x, y;

		nChessID = position[i][j];
		switch (nChessID) {
		case 55:
		case 5:

			for (y = 0; y < 3; y++)
				for (x = 3; x < 6; x++)
					if (canTouch(position, j, i, x, y))
						AddPoint(x, y);
			for (y = 7; y < 10; y++)
				for (x = 3; x < 6; x++)
					if (canTouch(position, j, i, x, y))
						AddPoint(x, y);
			break;

		case 44:

			for (y = 7; y < 10; y++)
				for (x = 3; x < 6; x++)
					if (canTouch(position, j, i, x, y))
						AddPoint(x, y);
			break;

		case 4:

			for (y = 0; y < 3; y++)
				for (x = 3; x < 6; x++)
					if (canTouch(position, j, i, x, y))
						AddPoint(x, y);
			break;

		case 33:
		case 3:

			x = j + 2;
			y = i + 2;
			if (x < 9 && y < 10 && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j + 2;
			y = i - 2;
			if (x < 9 && y >= 0 && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j - 2;
			y = i + 2;
			if (x >= 0 && y < 10 && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j - 2;
			y = i - 2;
			if (x >= 0 && y >= 0 && canTouch(position, j, i, x, y))
				AddPoint(x, y);
			break;

		case 22:
		case 2:
			x = j + 2;
			y = i + 1;
			if ((x < 9 && y < 10) && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j + 2;
			y = i - 1;
			if ((x < 9 && y >= 0) && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j - 2;
			y = i + 1;
			if ((x >= 0 && y < 10) && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j - 2;
			y = i - 1;
			if ((x >= 0 && y >= 0) && canTouch(position, j, i, x, y))
				AddPoint(x, y);

			x = j + 1;
			y = i + 2;
			if ((x < 9 && y < 10) && canTouch(position, j, i, x, y))
				AddPoint(x, y);
			x = j - 1;
			y = i + 2;
			if ((x >= 0 && y < 10) && canTouch(position, j, i, x, y))
				AddPoint(x, y);
			x = j + 1;
			y = i - 2;
			if ((x < 9 && y >= 0) && canTouch(position, j, i, x, y))
				AddPoint(x, y);
			x = j - 1;
			y = i - 2;
			if ((x >= 0 && y >= 0) && canTouch(position, j, i, x, y))
				AddPoint(x, y);
			break;

		case 11:
		case 1:
			x = j + 1;
			y = i;
			while (x < 9) {
				if (0 == position[y][x])
					AddPoint(x, y);
				else {
					AddPoint(x, y);
					break;
				}
				x++;
			}

			x = j - 1;
			y = i;
			while (x >= 0) {
				if (0 == position[y][x])
					AddPoint(x, y);
				else {
					AddPoint(x, y);
					break;
				}
				x--;
			}

			x = j;
			y = i + 1;//
			while (y < 10) {
				if (0 == position[y][x])
					AddPoint(x, y);
				else {
					AddPoint(x, y);
					break;
				}
				y++;
			}

			x = j;
			y = i - 1;// K
			while (y >= 0) {
				if (0 == position[y][x])
					AddPoint(x, y);
				else {
					AddPoint(x, y);
					break;
				}
				y--;
			}
			break;

		case 77:
			y = i - 1;
			x = j;

			if (y >= 0)
				AddPoint(x, y);

			if (i < 5) {
				y = i;
				x = j + 1;
				if (x < 9)
					AddPoint(x, y);
				x = j - 1;
				if (x >= 0)
					AddPoint(x, y);
			}
			break;

		case 7:
			y = i + 1;
			x = j;

			if (y < 10)
				AddPoint(x, y);

			if (i > 4) {
				y = i;
				x = j + 1;
				if (x < 9)
					AddPoint(x, y);
				x = j - 1;
				if (x >= 0)
					AddPoint(x, y);
			}
			break;

		case 66:
		case 6:

			x = j + 1; // T
			y = i;
			flag = false;
			while (x < 9) {
				if (0 == position[y][x]) {
					if (!flag)
						AddPoint(x, y);
				} else {
					if (!flag)
						flag = true;
					else {
						AddPoint(x, y);
						break;
					}
				}
				x++;
			}

			x = j - 1;
			flag = false;
			while (x >= 0) {
				if (0 == position[y][x]) {
					if (!flag)
						AddPoint(x, y);
				} else {
					if (!flag)
						flag = true;
					else {
						AddPoint(x, y);
						break;
					}
				}
				x--;
			}
			x = j;
			y = i + 1;
			flag = false;
			while (y < 10) {
				if (0 == position[y][x]) {
					if (!flag)
						AddPoint(x, y);
				} else {
					if (!flag)
						flag = true;
					else {
						AddPoint(x, y);
						break;
					}
				}
				y++;
			}

			y = i - 1; // K
			flag = false;
			while (y >= 0) {
				if (0 == position[y][x]) {
					if (!flag)
						AddPoint(x, y);
				} else {
					if (!flag)
						flag = true;
					else {
						AddPoint(x, y);
						break;
					}
				}
				y--;
			}
			break;

		default:
			break;

		}
		return nPosCount;
	}

	// judge if the a piece can be moved from A to B
	// If OK, return True,else return False
	private boolean canTouch(int[][] position, int nFromX, int nFromY, int nToX, int nToY) {
		int i = 0, j = 0;
		int nMoveChessID, nTargetID;

		if (nFromY == nToY && nFromX == nToX)
			return false;// ĿԴͬ

		nMoveChessID = position[nFromY][nFromX];
		nTargetID = position[nToY][nToX];

		switch (nMoveChessID) {
		case 5:
			if (nTargetID == 55)// Ͻ?
			{
				if (nFromX != nToX)
					return false;
				for (i = nFromY + 1; i < nToY; i++)
					if (position[i][nFromX] != 0)
						return false;
			} else {
				if (nToY > 2 || nToX > 5 || nToX < 3)
					return false;// ĿھŹ֮
				if (Math.abs(nFromY - nToY) + Math.abs(nToX - nFromX) > 1)
					return false;// ˧ֻһֱ:
			}
			break;
		case 44:

			if (nToY < 7 || nToX > 5 || nToX < 3)
				return false;// ʿŹ

			if (Math.abs(nFromY - nToY) != 1 || Math.abs(nToX - nFromX) != 1)
				return false; // ʿб

			break;

		case 4: // ʿ

			if (nToY > 2 || nToX > 5 || nToX < 3)
				return false;// ʿŹ

			if (Math.abs(nFromY - nToY) != 1 || Math.abs(nToX - nFromX) != 1)
				return false; // ʿб

			break;

		case 33:// 

			if (nToY < 5)
				return false;// ಻ܹ

			if (Math.abs(nFromX - nToX) != 2 || Math.abs(nFromY - nToY) != 2)
				return false;// 

			if (position[(nFromY + nToY) / 2][(nFromX + nToX) / 2] != 0)
				return false;// ۱ס

			break;

		case 3:// 

			if (nToY > 4)
				return false;// ಻ܹ

			if (Math.abs(nFromX - nToX) != 2 || Math.abs(nFromY - nToY) != 2)
				return false;// 

			if (position[(nFromY + nToY) / 2][(nFromX + nToX) / 2] != 0)
				return false;// ۱ס

			break;

		case 7: // ڱ

			if (nToY < nFromY)
				return false;// ͷ

			if (nFromY < 5 && nFromY == nToY)
				return false;// ǰֱֻ

			if (nToY - nFromY + Math.abs(nToX - nFromX) > 1)
				return false;// ֻһֱ:

			break;

		case 77: // 

			if (nToY > nFromY)
				return false;// ͷ

			if (nFromY > 4 && nFromY == nToY)
				return false;// ǰֱֻ

			if (nFromY - nToY + Math.abs(nToX - nFromX) > 1)
				return false;// ֻһֱ:

			break;

		case 55:
			if (nTargetID == 5)// Ͻ?
			{
				if (nFromX != nToX)
					return false;// ͬһ
				for (i = nFromY - 1; i > nToY; i--)
					if (position[i][nFromX] != 0)
						return false;// мб
			} else {
				if (nToY < 7 || nToX > 5 || nToX < 3)
					return false;// ĿھŹ֮
				if (Math.abs(nFromY - nToY) + Math.abs(nToX - nFromX) > 1)
					return false;// ˧ֻһֱ:
			}
			break;

		case 11:
		case 1:

			if (nFromY != nToY && nFromX != nToX)
				return false; // ֱ:

			if (nFromY == nToY) {
				if (nFromX < nToX) {
					for (i = nFromX + 1; i < nToX; i++)
						if (position[nFromY][i] != 0)
							return false;
				} else {
					for (i = nToX + 1; i < nFromX; i++)
						if (position[nFromY][i] != 0)
							return false;
				}
			} else {
				if (nFromY < nToY) {
					for (j = nFromY + 1; j < nToY; j++)
						if (position[j][nFromX] != 0)
							return false;
				} else {
					for (j = nToY + 1; j < nFromY; j++)
						if (position[j][nFromX] != 0)
							return false;
				}
			}

			break;

		case 22:
		case 2:

			if (!((Math.abs(nToX - nFromX) == 1 && Math.abs(nToY - nFromY) == 2)
					|| (Math.abs(nToX - nFromX) == 2 && Math.abs(nToY - nFromY) == 1)))
				return false;// 

			if (nToX - nFromX == 2) {
				i = nFromX + 1;
				j = nFromY;
			} else if (nFromX - nToX == 2) {
				i = nFromX - 1;
				j = nFromY;
			} else if (nToY - nFromY == 2) {
				i = nFromX;
				j = nFromY + 1;
			} else if (nFromY - nToY == 2) {
				i = nFromX;
				j = nFromY - 1;
			}

			if (position[j][i] != (int) 0)
				return false;// 

			break;
		case 66:
		case 6:

			if (nFromY != nToY && nFromX != nToX)
				return false; // ֱ:

			// ڲʱ·в:------------------

			if (position[nToY][nToX] == 0) {
				if (nFromY == nToY) {
					if (nFromX < nToX) {
						for (i = nFromX + 1; i < nToX; i++)
							if (position[nFromY][i] != 0)
								return false;
					} else {
						for (i = nToX + 1; i < nFromX; i++)
							if (position[nFromY][i] != 0)
								return false;
					}
				} else {
					if (nFromY < nToY) {
						for (j = nFromY + 1; j < nToY; j++)
							if (position[j][nFromX] != 0)
								return false;
					} else {
						for (j = nToY + 1; j < nFromY; j++)
							if (position[j][nFromX] != 0)
								return false;
					}
				}
			}
			// ڲ-------------------------------------
			// ʱ:=======================================
			else {
				int nCount = 0;
				if (nFromY == nToY) {
					if (nFromX < nToX) {
						for (i = nFromX + 1; i < nToX; i++)
							if (position[nFromY][i] != 0)
								nCount++;
						if (nCount != 1)
							return false;
					} else {
						for (i = nToX + 1; i < nFromX; i++)
							if (position[nFromY][i] != 0)
								nCount++;
						if (nCount != 1)
							return false;
					}
				} else {
					if (nFromY < nToY) {
						for (j = nFromY + 1; j < nToY; j++)
							if (position[j][nFromX] != 0)
								nCount++;
						if (nCount != 1)
							return false;
					} else {
						for (j = nToY + 1; j < nFromY; j++)
							if (position[j][nFromX] != 0)
								nCount++;
						if (nCount != 1)
							return false;
					}
				}
			}
			// ڳʱ================================
			break;
		default:
			return false;
		}

		return true;
	}

	public boolean IsSameSide(int a, int b) {
		if (a == 0 || b == 0) {
			return false;
		} else if (a < 10 && b > 10) {
			return false;
		} else if (a > 10 && b < 10) {
			return false;
		} else {
			return true;
		}
	}
	
	public int getCount() {
		return count;
	}

	public void setCount(int count) {
		this.count = count;
	}

}
