package ai;

public class Robot {
	protected int[][] CurPosition = new int[10][9];
	protected Evaluation m_pEval = new Evaluation();
	protected static int m_nSearchDepth = 3;
	protected int m_nMaxDepth;
	protected String name;
	protected Listener listener;
	protected int[] bestMove = new int[4];
	int nodeSearchCount;
	protected static String algorithm;

	Robot(Listener a) {
		listener = a;
	}

	public void memcpy(int[][] des, int[][] src) {
		for (int i = 0; i < 10; i++) {
			for (int j = 0; j < 9; j++) {
				des[i][j] = src[i][j];
			}
		}
	}

	public void bestmove(int bestmove_x, int bestmove_y, int bestmove_i, int bestmove_j) {
		bestMove[0] = bestmove_x;
		bestMove[1] = bestmove_y;
		bestMove[2] = bestmove_i;
		bestMove[3] = bestmove_j;
	}

	public static void setAlgorithm(String algorithm, int depth) {
		Robot.algorithm = algorithm;
		Robot.m_nSearchDepth = depth;
	}

	public int[] SearchAGoodMove(int[][] position) {
		m_nMaxDepth = m_nSearchDepth;
		CurPosition = position;

		long beginTime = System.currentTimeMillis();

		switch (algorithm) {
		case "Negamax Search":
			System.out.println("Algorithm: " + algorithm);
			System.out.println("         Search-depth: " + m_nSearchDepth);
			negaMax(m_nMaxDepth);
			break;
		case "Alpha-beta Search":
			System.out.println("Algorithm: " + algorithm);
			System.out.println("         Search-depth: " + m_nSearchDepth);
			alphabeta(m_nMaxDepth, -20000, 20000);
			break;
		case "Fail-soft Alpha-beta Search":
			System.out.println("Algorithm: " + algorithm);
			System.out.println("         Search-depth: " + m_nSearchDepth);
			fAlphabeta(m_nMaxDepth, -20000, 20000);
			break;
		case "Aspiration Search":
			System.out.println("Algorithm: " + algorithm);
			System.out.println("         Search-depth: " + m_nSearchDepth);
			AspirationSearch(position);
			break;
		case "Minimal Window Search":
			System.out.println("Algorithm: " + algorithm);
			System.out.println("         Search-depth: " + m_nSearchDepth);
			PVS(m_nMaxDepth, -20000, 20000);
			break;
		case "MTD(f)":
			int firstguess = 0;
			System.out.println("Algorithm: " + algorithm);
			System.out.println("         Search-depth: " + m_nSearchDepth);
			firstguess = MTDf(firstguess, m_nMaxDepth);
			break;
		}

		long endTime = System.currentTimeMillis();
		System.out.println("         Time-spent: " + (endTime - beginTime) + "ms");
		update(bestMove[0], bestMove[1], bestMove[2], bestMove[3]);
		return bestMove;
	}

	// Aspiration Search
	public int[] AspirationSearch(int[][] position) {
		int x, y;

		CurPosition = position;
		m_nMaxDepth = m_nSearchDepth - 1;
		x = fAlphabeta(m_nMaxDepth, -20000, 20000);

		m_nMaxDepth = m_nSearchDepth;
		y = fAlphabeta(m_nMaxDepth, x - 50, x + 50);

		if (y < x - 50) {// fail-low research
			fAlphabeta(m_nMaxDepth, -20000, y);
		}
		if (y > x + 50) {// fail-high research
			fAlphabeta(m_nMaxDepth, y, 20000);
		}

		update(bestMove[0], bestMove[1], bestMove[2], bestMove[3]);
		return bestMove;
	}

	protected void update(int rBefore, int cBefore, int rCur, int cCur) {
		listener.setcr(rCur, cCur);
		listener.setcur(rCur, cCur);
		listener.setbefore(rBefore, cBefore);
	}

	public boolean IsSameSide(int a, int b) {
		if (a == 0 || b == 0) {
			return false;
		} else if (a < 10 && b > 10) {
			return false;
		} else if (a > 10 && b < 10) {
			return false;
		} else {
			return true;
		}
	}

	protected int negaMax(int depth) {
		int current = -20000;
		int score;
		int i, j;

		nodeSearchCount++; // record the search count of leaf nodes

		i = isGameOver(CurPosition, depth); // check whether the game is over
		if (i != 0)
			return i;// when the game is over, return maximum/minimum value

		if (depth <= 0) // obtain the evaluation value of leaf node
			return m_pEval.evaluate(CurPosition, (m_nMaxDepth - depth) % 2 == 1 ? true : false);

		for (i = 0; i < 10; i++) { // list all possible moves
			for (j = 0; j < 9; j++) {
				if (CurPosition[i][j] == 0 || ((m_nMaxDepth - depth) % 2 == 0 && CurPosition[i][j] > 10)
						|| ((m_nMaxDepth - depth) % 2 == 1 && CurPosition[i][j] < 10)) {
					continue;
				}
				for (int x = 0; x < 10; x++) {
					for (int y = 0; y < 9; y++) {
						update(i, j, x, y);
						if (x == i && y == j) {
							continue;
						}
						if (IsSameSide(CurPosition[x][y], CurPosition[i][j])) {
							continue;
						}
						if (listener.ifwalk(CurPosition[i][j]) == 0) {
							continue;
						}
						int[][] temp = new int[10][9];
						memcpy(temp, CurPosition);// copy the current move
						listener.walk(true);
						score = -negaMax(depth - 1);
						memcpy(CurPosition, temp);// recover the current move

						if (score > current) {
							current = score;
							if (depth == m_nMaxDepth) { // save the best move
								bestmove(x, y, i, j);
							}
						}
					}
				}

			}

		}
		return current; // return the maximum value
	}

	protected int alphabeta(int depth, int alpha, int beta) {
		int score;
		int i, j;

		nodeSearchCount++;

		i = isGameOver(CurPosition, depth);
		if (i != 0)
			return i;

		if (depth <= 0) // obtain the evaluation value of leaf node
			return m_pEval.evaluate(CurPosition, (m_nMaxDepth - depth) % 2 == 1 ? true : false);

		for (i = 0; i < 10; i++) {
			for (j = 0; j < 9; j++) {
				if (CurPosition[i][j] == 0 || ((m_nMaxDepth - depth) % 2 == 0 && CurPosition[i][j] > 10)
						|| ((m_nMaxDepth - depth) % 2 == 1 && CurPosition[i][j] < 10)) {
					continue;
				}
				for (int x = 0; x < 10; x++) {
					for (int y = 0; y < 9; y++) {
						update(i, j, x, y);
						if (x == i && y == j) {
							continue;
						}
						if (IsSameSide(CurPosition[x][y], CurPosition[i][j])) {
							continue;
						}
						if (listener.ifwalk(CurPosition[i][j]) == 0) {
							continue;
						}
						int[][] temp = new int[10][9];
						memcpy(temp, CurPosition);
						listener.walk(true);
						score = -alphabeta(depth - 1, -beta, -alpha);
						memcpy(CurPosition, temp);

						if (score > alpha) {
							alpha = score;// retain the maximum value
							// record the best move when closing to the root node
							if (depth == m_nMaxDepth) {
								bestmove(x, y, i, j);
							}
						}
						if (alpha >= beta) {
							break;// prune and give the rest of nodes
						}
					}
				}

			}

		}
		return alpha;// return maximum value
	}

	protected int fAlphabeta(int depth, int alpha, int beta) {
		int current = -20000;
		int score;
		int i, j;

		i = isGameOver(CurPosition, depth);
		if (i != 0)
			return i;

		if (depth <= 0) // obtain the evaluation value of leaf node
			return m_pEval.evaluate(CurPosition, (m_nMaxDepth - depth) % 2 == 1 ? true : false);

		for (i = 0; i < 10; i++) {
			for (j = 0; j < 9; j++) {
				if (CurPosition[i][j] == 0 || ((m_nMaxDepth - depth) % 2 == 0 && CurPosition[i][j] > 10)
						|| ((m_nMaxDepth - depth) % 2 == 1 && CurPosition[i][j] < 10)) {
					continue;
				}
				for (int x = 0; x < 10; x++) {
					for (int y = 0; y < 9; y++) {
						update(i, j, x, y);
						if (x == i && y == j) {
							continue;
						}
						if (IsSameSide(CurPosition[x][y], CurPosition[i][j])) {
							continue;
						}
						if (listener.ifwalk(CurPosition[i][j]) == 0) {
							continue;
						}
						int[][] temp = new int[10][9];
						memcpy(temp, CurPosition);
						listener.walk(true);
						score = -alphabeta(depth - 1, -beta, -alpha);
						memcpy(CurPosition, temp);

						if (score > current) {
							current = score;// retain the maximum value

							// // record the best move when closing to the root node
							if (depth == m_nMaxDepth) {
								bestmove(x, y, i, j);
							}
							if (score > alpha)
								alpha = score;// modify the boundary of alpha
							if (score >= beta)
								break;// beta pruning
						}
					}
				}

			}

		}

		return current;
	}

	protected int PVS(int depth, int alpha, int beta) {
		int current = 0, score = 0;
		int i, j;
		boolean index = false;
		nodeSearchCount++;

		i = isGameOver(CurPosition, depth);
		if (i != 0)
			return i;
		if (depth <= 0) // obtain the evaluation value of leaf node
			return m_pEval.evaluate(CurPosition, (m_nMaxDepth - depth) % 2 == 1 ? true : false);
		for (i = 0; i < 10; i++) {
			for (j = 0; j < 9; j++) {
				if (CurPosition[i][j] == 0 || ((m_nMaxDepth - depth) % 2 == 0 && CurPosition[i][j] > 10)
						|| ((m_nMaxDepth - depth) % 2 == 1 && CurPosition[i][j] < 10)) {
					continue;
				}
				for (int x = 0; x < 10; x++) {
					for (int y = 0; y < 9; y++) {
						update(i, j, x, y);
						if (x == i && y == j) {
							continue;
						}
						if (IsSameSide(CurPosition[x][y], CurPosition[i][j])) {
							continue;
						}
						if (listener.ifwalk(CurPosition[i][j]) == 0) {
							continue;
						}

						if (!index) {
							int[][] temp1 = new int[10][9];
							memcpy(temp1, CurPosition);
							listener.walk(true);
							current = -PVS(depth - 1, -beta, -alpha);
							memcpy(CurPosition, temp1);
							if (depth == m_nMaxDepth) {
								bestmove(x, y, i, j);
							}
							index = true;
						} else {
							if (current < beta) { // if cannot beta pruning
								if (current > alpha) {
									alpha = current;
								}
								int[][] temp1 = new int[10][9];
								memcpy(temp1, CurPosition);
								listener.walk(true);
								score = -PVS(depth - 1, -alpha - 1, -alpha);// narrow window search
								if ((score > alpha) && (score < beta)) {
									current = -PVS(depth - 1, -beta, -score);// fail-high, search again
									if (depth == m_nMaxDepth) {
										bestmove(x, y, i, j);
									}
								} else if (score > current) {
									current = score; // narrow window search hits
									if (depth == m_nMaxDepth) {
										bestmove(x, y, i, j);
									}
								}
								memcpy(CurPosition, temp1);
							}
						}

					}
				}

			}

		}
		return current; // return the best value
	}

	protected int MTDf(int firstguess, int depth) {
		int score, lowerbound, upperbound, beta;
		score = firstguess;
		// the initial search range is between -20000 and 20000
		upperbound = 20000;
		lowerbound = -20000;
		while (lowerbound < upperbound) {
			// move the window to the object
			if (score == lowerbound) {
				beta = score + 1;
			} else {
				beta = score;
			}
			score = PVS(depth, beta - 1, beta);// Empty window search
			if (score < beta) {
				upperbound = score;
			} else {
				lowerbound = score;
			}
		}
		return score;
	}

	protected int isGameOver(int[][] position, int nDepth) {
		int i, j;
		boolean RedLive = false, BlackLive = false;

		for (i = 7; i < 10; i++)
			for (j = 3; j < 6; j++) {
				if (position[i][j] == 5)
					BlackLive = true;
				if (position[i][j] == 55)
					RedLive = true;
			}

		for (i = 0; i < 3; i++)
			for (j = 3; j < 6; j++) {
				if (position[i][j] == 5)
					BlackLive = true;
				if (position[i][j] == 55)
					RedLive = true;
			}

		i = (m_nMaxDepth - nDepth + 1) % 2;// һ˾Ϊ¹γɵľ棬öһ򲻼һ

		if (!RedLive) // 콫Ƿ
			if (i != 0)
				return 19990 + nDepth;// 㣨к--ں֣עֵǸֲһ1һҪӣһºһǰؼֵ;
			else
				return -19990 - nDepth;// ż㣨к󣩷ؼСֵ
		if (!BlackLive)
			if (i != 0)
				return -19990 - nDepth;// 㣨к󣩷ؼСֵ;
			else
				return 19990 + nDepth;// ż㣨к󣩷ؼֵ;
		return 0;
	}
}
